import scala.actors._
import scala.actors.Actor._

case class User(name: String)
case class Subscribe(user: User)
case class Unsubscribe(user: User)
case class Post(msg: String)
case class PostFrom(user: User, post: Post)

class ChatRoom extends Actor {

  def act() {
    var session = Map.empty[User, Actor]
    while (true) {
      receive {
        case Subscribe(user) =>
		  Thread.sleep(10 * 1000) // Introduce latency in the message
		  						  // processing
          val subscriberUser = actor {
            while (true) {
              self.receiveWithin(10 * 1000) {
                case Post(msg) =>
                  println(msg)
				case TIMEOUT =>
				   this ! Unsubscribe(user)
				   exit()
              }
            }
          }

          session = session + (user -> subscriberUser)
		  reply("Subscribed " + user)

        case PostFrom(user, msg) =>
          for (k <- session.keys) {
            session(k) ! msg
          }

		case Unsubscribe(user) =>
		  if (session.contains(user)) {
		    session  -= user
		    println("Unsubscribed " + user)
		  }
      }
    }
  }
}

object FuturesChat {
  def main(args: Array[String]) {
	require(args.length == 1)
	val user = args(0)

    val chatroom = new ChatRoom
    chatroom.start()

	// The !! (futures) operator does not block the current thread.
	// Notice that we introduce 10 seconds of latency for the chatroom
	// to send back a reply. Even though the current thread is not
	// blocked and we can do useful work, the reply is still not available. 
	// We will obtain the reply by evaluating the response using the
	// invocation syntax ()
	println("Sending subscription request for " + user + "...")
    val response = chatroom !! Subscribe(User(user))
	println("Awaiting subscription response, but we can do something else meanwhile...")
	Thread.sleep(5000) // This simulates some useful work ... or just sleeping
	println("Feeling antsy now...") // Still no reply from the chat room

	// Evaluating the reply now causes the current thread to block until
	// we finally get what we're waiting for. There is also a way to
	// specify a timeout for the wait. 
	println("Finally! " + response() + " Feel free to post away.")

	// We are now ready to post messages to the chatroom
	while (true)  {
		val r = readLine
    	chatroom ! PostFrom(User(user), Post(r))
	}
  }
}
