/*
A chat application that automatically unsubscribes a user if no messages
are received within 10 seconds. 
*/
import scala.actors._
import scala.actors.Actor._

case class User(name: String)
case class Subscribe(user: User)
case class Unsubscribe(user: User)
case class Post(msg: String)
case class PostFrom(user: User, post: Post)

class ChatRoom extends Actor {

  def act() {
    var session = Map.empty[User, Actor]
    while (true) {
      receive {
        case Subscribe(user) =>
          val subscriberUser = actor {
            while (true) {
			  // Using receiveWithin will cause a TIMEOUT messages to be
			  // sent to this Actor if no message was received in the
			  // mailbox within 10 seconds.
              self.receiveWithin(10 * 1000) {
                case Post(msg) => println(msg)

				case TIMEOUT =>
				   this ! Unsubscribe(user) // this refers to the
				   							// "outer" Actor

				   exit() // exit() causes the Actor to terminate 
				   		  // message processing
              }
            }
          }

          session = session + (user -> subscriberUser)
		  println("Subscribed " + user)

        case PostFrom(user, msg) =>
          for (k <- session.keys) {
            session(k) ! msg
          }

		// Process the Unsubscribe message by removing the user from the
		// session. This will make the subscriberUser eligible for GC
		case Unsubscribe(user) =>
		  if (session.contains(user)) {
		    session  -= user
		    println("Unsubscribed " + user)
		  }
      }
    }
  }
}

object ReceiveWithinChat {
  def main(args: Array[String]) {
	require(args.length == 1)
	val user = args(0)

    val chatroom = new ChatRoom
    chatroom.start()
    chatroom ! Subscribe(User(user))

	while (true)  {
		val r = readLine
    	chatroom ! PostFrom(User(user), Post(r))
	}
  }
}
