/* 
 * A remote chat application.
 *
 * Start server and client processes on 3 different terminals:
 * 
 * scala RemoteChatServer
 *
 * scala RemoteChatClient alice 8001
 *
 * scala RemoteChatClient bob 8002
 */
import scala.actors._
import scala.actors.Actor._
import scala.actors.remote.Node
import scala.actors.remote.RemoteActor._

case class User(name: String)
case class Subscribe(user: User)
case class Unsubscribe(user: User)
case class Post(msg: String)
case class PostFrom(user: User, post: Post)
case class Repost(user: User, post: Post, channel: OutputChannel[Any])

class ChatRoom extends Actor {

 Debug.level = 9

  def act() {
	alive(8000) 				// Listen for messages on port 8000
	register('chatRoom, self)	// Register for lookup 

    var session = Map.empty[User, Actor]
    while (true) {
      receive {
        case Subscribe(user) =>
		  val subscriber = sender // Capture the sender in variable

          val sessionUser = actor {
            while (true) {
              self.receiveWithin(90 * 1000) {

				// Process a Post message. 
				// msg captures the Post(m) value
				// Note that subscriber is available, because 
				// actor{} defines a closure. So the value of 
				// subscriber assigned previously is captured in this
				// closure
                case msg @ Post(m) => subscriber ! msg

				case TIMEOUT =>
				   this ! Unsubscribe(user)
				   exit()
              }
            }
          }
          session = session + (user -> sessionUser)
          println("Subscribed " + user)

        case PostFrom(user, post) =>
			// Don't the message back to the poster
          	for (k <- session.keys; if k != user) {
              	session(k) ! Post(user.name + " says: " + post.msg)
          	}

		case Unsubscribe(user) =>
		  if (session.contains(user)) {
		    println("Unsubscribed " + user)
			val actr = session(user)
		    session  -= user
		  }
      }
    }
  }
}

object RemoteChatServer {
  def main(args: Array[String]) {
    val chatroom = new ChatRoom
    chatroom.start()
	println("Chat server running")
  }
}


/*
 * The client app is a singleton object that consumes two parameters from
 * the command line: the name of the user to register as, and a port
 * number to make the client active on. The chat server will use that
 * port to send messages back to the client.
 *
 * When the client starts up, it attempts to connect to the chat server,
 * and register the user with it. You can can then type messages in the
 * command line, and the chat server will distribute those messages to
 * other clients. Messages from other user, in turn, are delivered to
 * this client. 
 * 
 * This example assumes that the server process listens on port 8000 of the same
 * host as the client.
 */

object RemoteChatClient {

  def main(args: Array[String]) {

    // Incomment this to receive lots of debug messages
	//Debug.level = 99

	require(args.length == 2)
	val chatroom = select(Node("localhost", 8000), 'chatRoom)
	val user = args(0)
	var port = args(1).toInt
	val actr = actor {
		alive(port)
		register('aClient, self)
		println("Registered on " + port)
		chatroom ! Subscribe(User(user))
		while (true) {
			receive {
				case Post(msg) => println(msg)
			}
		}
	}
	while (true)  {
	    val r = readLine
        chatroom ! PostFrom(User(user), Post(r))
	}
  }
}
